﻿using System;
using System.Collections.Generic;
using System.Windows.Forms;


/* Partial fraction decomposition */
/* www.mosismath.com              */


namespace PartialFraction
{
    public enum POLETYPE
    {
        Single = 0,
        Multiple = 1,
        Complex = 2
    }
     
    public partial class Mainwin : Form
    {
        int orderDenom = 6;
        PartialFraction[] partialFract;
        double[] enumeratorPolynom;
        double[] denominatorPolynom;  // input polynomial
        public List<double> realRoots = new List<double>();
        public QRAlgorithm qr;

        public Mainwin()
        {
            int i;
            DataGridViewCell cell;

            InitializeComponent();

            partialFract = new PartialFraction[0];
            enumeratorPolynom = new double[orderDenom];
           
            denominatorPolynom = new double[orderDenom];  // polynomial
            qr = new QRAlgorithm(orderDenom);

            dgPolinomial.ColumnCount = orderDenom;
            dgPolinomial.RowCount = 2;

            sgSolution.ColumnCount = orderDenom;
            sgSolution.RowCount = 2;

            dGMatrix.ColumnCount = orderDenom;
            dGMatrix.RowCount = orderDenom;

            dgResult.ColumnCount = 1;
            dgResult.RowCount = orderDenom;

            cbOrderDenom.SelectedIndex = 3;

            denominatorPolynom[0] = 1.0;  // (x + 1)*(x^2 + 6x + 13)^2
            denominatorPolynom[1] = 13.0; 
            denominatorPolynom[2] = 74.0;
            denominatorPolynom[3] = 218.0;
            denominatorPolynom[4] = 325.0;
            denominatorPolynom[5] = 169.0;

            enumeratorPolynom[0] = 2.0;
            enumeratorPolynom[1] = 28.0;
            enumeratorPolynom[2] = 162.0;
            enumeratorPolynom[3] = 440.0;
            enumeratorPolynom[4] = 496.0;

            for (i = 0; i < enumeratorPolynom.Length; i++)
            {
                cell = dgPolinomial[i, 0];  // get cell to access
                cell.Value = Convert.ToString(enumeratorPolynom[i]);
            }

            for (i = 0; i < denominatorPolynom.Length; i++)
            {
                cell = dgPolinomial[i, 1];  // get cell to access
                cell.Value = Convert.ToString(denominatorPolynom[i]);
            }
        }

        public double[] MultPoly(double[] p1, double[] p2)
        {
            int i, j;
            double[] tempPoly = new double[p1.Length + p2.Length - 1];
            for (i = 0; i < p1.Length; i++)
            {
                for (j = 0; j < p2.Length; j++)
                {
                    tempPoly[j + i] = tempPoly[j + i] + p2[j] * p1[i];
                }
            }
            return tempPoly;
        }

        public double[] MultPoly(double[] p1, double p2)
        {
            int i;
            for (i = 0; i < p1.Length; i++)
            {
                p1[i] = p1[i] * p2;
            }
            return p1;
        }


        public void CopyPoly(double[] source, ref double[] dest)
        {
            int i;
            Array.Resize(ref dest, source.Length);
            for (i = 0; i < dest.Length; i++)
            {
                dest[i] = source[i];
            }
        }


        private bool AreEqual(double[] a, double[] b)
        {
            bool equal = true;
            int i;
            if (a.Length == b.Length)
            {
                for (i = 0; i < a.Length; i++)
                {
                    if (Math.Abs(a[i] - b[i]) > 1E-3)
                        equal = false;
                }
            }
            else
                equal = false;
            return equal;
        }


        private void BuildMatrixAndSolve(int numberOfFractions)
        {
            int i, j, k;
            int actPosition = 0;
            DataGridViewCell cell;
            double[] equationValue = new double[numberOfFractions];
            TMatrix m = new TMatrix(numberOfFractions);
            TGauss gauss = new TGauss(numberOfFractions, m);
            double[] tempDenom = new double[1];
            double nom;
            double res;
            double s = -5;

            for (i = 0; i < numberOfFractions; i++)
            {
                while (realRoots.Contains(s))
                    s++;
                equationValue[i] = s;
                s++;
            }

            for (k = 0; k < numberOfFractions; k++)
            {
                for (i = 0; i < numberOfFractions; i++)
                {
                    nom = 0;
                    Array.Resize(ref tempDenom, actPosition + 1);
                    tempDenom[0] = 1;
                    for (j = 0; j < numberOfFractions; j++)
                    {
                        if ((j != i) && !((partialFract[i].type > POLETYPE.Single) && (j == partialFract[i].parent)))
                            tempDenom = MultPoly(tempDenom, partialFract[j].expander);
                    }
                    if (partialFract[i].type > POLETYPE.Single)
                    {
                        tempDenom = MultPoly(tempDenom, partialFract[i].multiplier);
                    }
                    for (j = 0; j < tempDenom.Length; j++)
                    {
                        nom = nom + tempDenom[j] * Math.Pow(equationValue[k], tempDenom.Length - j - 1);
                    }
                    cell = dGMatrix[i, k];
                    cell.Value = Convert.ToString(Convert.ToString(nom));
                    m.a[k, i] = nom;
                }
                res = 0;
                for (j = 0; j < enumeratorPolynom.Length; j++)
                {
                    res = res + enumeratorPolynom[j] * Math.Pow(equationValue[k], enumeratorPolynom.Length - j - 1);
                }
                cell = dgResult[0, k];
                cell.Value = Convert.ToString(Convert.ToString(res));
                m.y[k] = res;
            }
            gauss.Eliminate();
            gauss.Solve();

            for (i = 0; i < numberOfFractions; i++)
            {
                cell = sgSolution[i, 0];
                cell.Value = WritePolynom(MultPoly(partialFract[i].enumerator, m.x[i]),1);
                cell = sgSolution[i, 1];
                cell.Value = WritePolynom(partialFract[i].denominator, partialFract[i].power);
            }
        }

       
        private void btCalc_Click(object sender, EventArgs e)
        {
            int i, k;
            int actOrderQR = orderDenom - 1;
            int enumlength = 0;
            bool found = false;
            bool parseError = false;
            Cursor.Current = Cursors.WaitCursor;
            DataGridViewCell cell;
            Array.Resize(ref enumeratorPolynom, orderDenom);

            // clear data grids
            for (i = 0; i < dGMatrix.ColumnCount; i++)
            {
                for (k = 0; k < dGMatrix.RowCount; k++)
                {
                    cell = dGMatrix[i, k];
                    cell.Value ="";
                }
            }
            for (k = 0; k < dgResult.RowCount; k++)
            {
                cell = dgResult[0, k];
                cell.Value = "";
            }
            for (i = 0; i < sgSolution.ColumnCount; i++)
            {
                for (k = 0; k < sgSolution.RowCount; k++)
                {
                    cell = sgSolution[i, k];
                    cell.Value = "";
                }
            }

            cell = dgPolinomial[dgPolinomial.ColumnCount-1, 1];  // get cell to access
            if ((cell.Value == null) || (cell.Value.ToString() == "0"))
            {
                MessageBox.Show("Last value should not be 0", " Enumerator polynomial");
                parseError = true;
            }

            // read data from datagrids
            for (i = 0; i < dgPolinomial.ColumnCount; i++)
            {
                if (i < enumeratorPolynom.Length)
                {
                    cell = dgPolinomial[i, 0];  // get cell to access
                    if (cell.Value != null)
                    {
                        try
                        {
                            enumeratorPolynom[i] = Convert.ToDouble(cell.Value);
                        }
                        catch
                        {
                            MessageBox.Show("No valid number", " Enumerator polynomial");
                            enumeratorPolynom[i] = 0;
                            parseError = true;
                        }
                    }
                    else
                        enumeratorPolynom[i] = 0;
                }

                if (i < denominatorPolynom.Length)
                {
                    cell = dgPolinomial[i, 1];  // get cell to access
                    if (cell.Value != null)
                    {
                        try
                        {
                            denominatorPolynom[i] = Convert.ToDouble(cell.Value);
                        }
                        catch
                        {
                            MessageBox.Show("No valid number", "Denominator polynomial");
                            enumeratorPolynom[i] = 0;
                            parseError = true;
                        }
                    }
                    else
                        denominatorPolynom[i] = 0;
                }
            }
            if (!parseError)
            {
                // adjust the length of the enumerator polynomial
                enumlength = enumeratorPolynom.Length;
                for (i = enumeratorPolynom.Length - 1; i >= 0; i--)
                {
                    if (enumeratorPolynom[i] == 0)
                        enumlength--;
                }
                Array.Resize(ref enumeratorPolynom, enumlength);

                for (i = 0; i < actOrderQR; i++)
                {
                    for (k = 0; k < actOrderQR; k++)
                    {
                        qr.m.a[k, i] = 0.0;

                    }
                    if (i < actOrderQR - 1)
                        qr.m.a[i + 1, i] = 1.0;
                }
                for (i = 0; i < actOrderQR; i++)
                    qr.m.a[i, actOrderQR - 1] = -denominatorPolynom[orderDenom - i - 1];

                qr.run();  // qr decomposation to compute the roots of the denominator polynomial

                Array.Resize(ref partialFract, qr.roots.Count);
                i = 0;

                foreach (double[] root in qr.roots)
                {
                    found = false;
                    if (i > 0)
                    {
                        for (k = 0; k < i; k++)
                        {
                            double[] tempRoot = new double[2];
                            CopyPoly(root, ref tempRoot);
                            if (AreEqual(partialFract[k].denominator, tempRoot) && !found)
                            {
                                // already one root of the same value found -> multiple root
                                if (!found)
                                {
                                    // allready one pole found => multiple real pole
                                    partialFract[i] = new PartialFraction(1);
                                    partialFract[i].expander[0] = 1.0;
                                    partialFract[i].InitDenominator(root);
                                    partialFract[i].multiplier = MultPoly(partialFract[i].multiplier, partialFract[k].expander);
                                    partialFract[i].parent = k;
                                    partialFract[i].type = POLETYPE.Multiple;
                                    partialFract[k].type = POLETYPE.Multiple;

                                    if (root.Length > 2) // complex root
                                    {
                                        partialFract[i].multiplier[0] = 1.0;
                                        Array.Resize(ref partialFract, partialFract.Length + 1);
                                        i++;

                                        partialFract[i] = new PartialFraction(1);
                                        partialFract[i].expander[0] = 1.0;
                                        partialFract[i].InitDenominator(root);
                                        Array.Resize(ref partialFract[i].enumerator, 2);
                                        partialFract[i].enumerator[1] = 0;
                                        Array.Resize(ref partialFract[i].multiplier, 2);
                                        partialFract[i].multiplier[0] = 1.0;
                                        partialFract[i].multiplier[1] = 0;
                                        partialFract[i].multiplier = MultPoly(partialFract[i].multiplier, partialFract[k].expander);
                                        partialFract[i].type = POLETYPE.Multiple;
                                        partialFract[i].parent = k;
                                        partialFract[k + 1].power++;
                                    }
                                    else
                                    {
                                        // real root
                                        realRoots.Add(Math.Round(root[1], 2));
                                    }
                                    partialFract[k].expander = MultPoly(partialFract[k].expander, tempRoot);
                                    partialFract[k].power++;
                                    found = true;
                                    i++;
                                }
                            }
                        }
                    }

                    if (!found) 
                    {
                        partialFract[i] = new PartialFraction(root.Length);
                        partialFract[i].InitExpander(root);
                        partialFract[i].InitDenominator(root);
                        if (root.Length > 2) // complex root
                        {
                            // conjugate complex pole pair
                            partialFract[i].type = POLETYPE.Complex;
                            partialFract[i].multiplier[0] = 1.0;
                            Array.Resize(ref partialFract, partialFract.Length + 1);
                            i++;

                            partialFract[i] = new PartialFraction(1);
                            partialFract[i].expander[0] = 1.0;
                            partialFract[i].InitDenominator(root);
                            Array.Resize(ref partialFract[i].enumerator, 2);
                            partialFract[i].enumerator[1] = 0;
                            Array.Resize(ref partialFract[i].multiplier, 2);
                            partialFract[i].multiplier[0] = 1.0;
                            partialFract[i].multiplier[1] = 0;
                            partialFract[i].type = POLETYPE.Complex;
                            partialFract[i].parent = i - 1;
                        }
                        else
                        {
                            // real root
                            partialFract[i].type = POLETYPE.Single;
                            realRoots.Add(Math.Round(root[1], 2));
                        }
                        i++;
                    }
                }
                SetPowerValues();
                BuildMatrixAndSolve(partialFract.Length);
            }
            Cursor.Current = Cursors.Default;
        }

        struct SPOWER
        {
            public int index;
            public int power;
        }

        private void SetPowerValues()
        {
            int i, j;
            SPOWER[] pwrValues = new SPOWER[0];
     
            for (i = 0; i < partialFract.Length; i++)
            {
                if (partialFract[i].power > 1)
                {
                    Array.Resize(ref pwrValues, pwrValues.Length + 1);
                    pwrValues[pwrValues.Length-1].index = i;
                    pwrValues[pwrValues.Length - 1].power = partialFract[i].power;
                }
                else
                {
                    if (pwrValues.Length >= 1)
                    {
                        for(j = 0; j < pwrValues.Length; j++)
                        {
                            if (pwrValues[j].index == partialFract[i].parent)
                            {
                                pwrValues[j].power--;
                                if (partialFract[i].denominator.Length > 2)
                                {
                                    partialFract[i].power = pwrValues[j].power;
                                    partialFract[i + 1].power = pwrValues[j].power;
                                    i++;
                                }
                                else
                                    partialFract[i].power = pwrValues[j].power;
                            }
                        }
                    }
                }
            }
        }


        private string WritePolynom(double[] polynomial, int power)
        {
            string tempStr = "";
            int j;
 
            for (j = 0; j < polynomial.Length; j++)
            {
                if (polynomial[j] != 0)
                {
                    if (tempStr != "")
                    {
                        if (polynomial.Length - j > 2)
                        {
                            if (polynomial[j] != 1.0)
                                tempStr = tempStr + " + " + Convert.ToString(Math.Round(polynomial[j], 3)) + "s^" + Convert.ToString(polynomial.Length - j - 1);
                            else
                                tempStr = tempStr + " + s^" + Convert.ToString(polynomial.Length - j - 1);
                        }
                        else
                        {
                            if (polynomial.Length - j > 1)
                            {
                                if (polynomial[j] != 1.0)
                                    tempStr = tempStr + " + " + Convert.ToString(Math.Round(polynomial[j], 3)) + "s";
                                else
                                    tempStr = tempStr + " + s";
                            }
                            else
                                tempStr = tempStr + " + " + Convert.ToString(Math.Round(polynomial[j],3));
                        }
                    }
                    else
                    {
                        if (polynomial.Length - j > 2)
                        {
                            if (polynomial[j] != 1.0)
                                tempStr = Convert.ToString(Math.Round(polynomial[j], 3)) + " s^" + Convert.ToString(polynomial.Length - j - 1);
                            else
                                tempStr = "s^" + Convert.ToString(polynomial.Length - j - 1);
                        }
                        else
                        {
                            if (polynomial.Length - j > 1)
                            {
                                if (polynomial[j] != 1.0)
                                    tempStr = Convert.ToString(Math.Round(polynomial[j], 3)) + "s";
                                else
                                    tempStr = "s";
                            }
                            else
                                tempStr = tempStr + Convert.ToString(Math.Round(polynomial[j], 3));
                        }
                    }
                }
            }
            if (power > 1)
                return "(" + tempStr + ")^" + power.ToString();
            else
                return tempStr;
        }

        private void cbOrderDenom_SelectedIndexChanged(object sender, EventArgs e)
        {
            orderDenom = cbOrderDenom.SelectedIndex + 3;

            Array.Resize(ref denominatorPolynom, orderDenom);
            Array.Resize(ref enumeratorPolynom, orderDenom);

            qr = new QRAlgorithm(orderDenom);
            dgPolinomial.ColumnCount = orderDenom; 
            sgSolution.ColumnCount = orderDenom;
            dGMatrix.ColumnCount = orderDenom;
            dGMatrix.RowCount = orderDenom;
            dgResult.RowCount = orderDenom;
        }
    }
}
